
# dobserv_funcs.r - Double-observer functions
.onAttach <- function(libname,pkgname)
  packageStartupMessage(pkgname,packageVersion(pkgname),
     paste0(' - for help and examples, type ?',pkgname))

logx <- function(x) {  x[x<=0]=1e-12; return(log(x))}

pool_lt10 <- function(dobsdata) {
  #' Pool species data with < 10 detections per species
  #'
  #' \code{pool_lt10} pools species w/ < 10 detections
  #'
  #'@param dobsdata input data-matrix with 4 columns:
  #' \itemize{
  #' \item{observerNumber - number of observer who is primary observer}
  #' \item{species - char string with species name}
  #' \item{N1 - number of individuals of species seen by primary observer}
  #' \item{N2 - number of individuals of species seen by secondary observer which were missed by primary observer}
  #' }
  #'
  #'@return data matrix w/ speciesName replaced if < 10 detections
  #'@export
  #'@author Jim Hines
  #'@examples
  #' \dontrun{
  #' fname=system.file('extdata','sample.csv',package='dobserv')
  #' data=read.csv(fname,header=F,skip=1)
  #' pooled_data=pool_lt10(data)
  #' }
  #'
  #  read input file and summarize number of detections by species...
  sp=spx=as.character(dobsdata[,2]);
  a1=aggregate(dobsdata[,3]+dobsdata[,4],by=list(dobsdata[,2]),sum)
  #  get species with less than 10 detections
  i=which(a1$x<10); sp9=as.character(a1$Group.1[i])
  #  for each species w/ < 10 detections, replace species name w/ 'grp1'
  for (j in 1:length(sp)) if (as.character(sp[j]) %in% sp9) spx[j]='grp1'
  return(data.frame(obs=dobsdata[,1],sp=spx,X1=dobsdata[,3],X01=dobsdata[,4]))
}

like_dep <- function(theta,d,dm) {
  #'  likelihood function for dependent observers
  #'
  #'  \code{like_dep} returns likelihood value for parameter vector theta,
  #'  data matrix d, and design matrix, dm
  #'
  #'@param theta vector of parameters (detection probs)
  #'@param d data matrix
  #'@param dm design matrix
  #'
  #'@return negative log-likelihood value
  #'
  #'@author Jim Hines
  #'
  #'@seealso like_indep

  th=plogis(dm %*% theta); nreal=nrow(dm);
  p1=th[seq(1,nreal,2)];
  p2=th[seq(2,nreal,2)]
  denom=p1+p2-p1*p2;
  prb1=cbind(p1/denom,p2/denom)
  i=cbind(1:nrow(d), 2-(d$obs==d$obs[1]))
  like=(d$X1*logx(prb1[i]) + d$X01*logx(1-prb1[i]))
  n1=d$X1+d$X01;
  expval<<-cbind(n1*prb1[i],n1*(1-prb1[i]))
  return(-sum(like))
}
like_indep <- function(theta,d,dm) {
  #'  likelihood function for independent observers
  #'
  #'  \code{like_indep} returns likelihood value for parameter vector theta,
  #'  data matrix d, and design matrix, dm
  #'
  #'@param theta vector of parameters (detection probs)
  #'@param d data matrix
  #'@param dm design matrix
  #'
  #'@return negative log-likelihood value
  #'
  #'@author Jim Hines
  #'
  #'@seealso like_indep
  #'

  th=plogis(dm %*% theta); nreal=nrow(dm);
  thj=th[seq(1,nreal,2)];
  thjp1=th[seq(2,nreal,2)]
  denom=1-(1-thj)*(1-thjp1)
  p1=thj*(1-thjp1)/denom;
  p2=(1-thj)*thjp1/denom
  like=sum(d[,1]*log(p1) + d[,2]*log(p2) + d[,3]*log(1-p1-p2))
  n1=rowSums(cbind(d$X10,d$X01,d$X11))
  expval<<-cbind(n1*p1,n1*p2,n1*(1-p1-p2))
  return(-like)
}
invhess <- function(h,est) {
  #'  invert hessian matrix
  #'
  #'  \code{invhess} inverts hessian matrix from optmiz function
  #'  deleting rows/cols of all zeros if necessary
  #'  and inserting zeros into result
  #'@param h input hessian matrix
  #'@param est input estimate vector
  #'@return inverse matrix
  #'
  i=which(rowSums(h)==0); h2=h3=matrix(0,nrow(h),ncol(h))
  i=which(abs(est)>15)
  if (length(i)==0) {
      res=try(h3<-solve(h))
  } else {
      res=try(h2<-solve(h[-i,-i]),T); k=0
      if (class(res)[1]!='try-error') {
        for (j in 1:ncol(h)) {
          if (!(j %in% i)) {
            k=k+1; a=rep(0,ncol(h)); a[-i]=h2[,k]; h3[,j]=a; h3[j,]=a
          }
        }
      }
  }
  return(h3)
}
runmod <- function(frm,mx,d,type='dependent') {
  #'  runs double-observer model
  #'
  #'  \code{runmod} run doble-obs model
  #'
  #'@param frm formula for model (eg., ~obs)
  #'@param mx species list
  #'@param d data frame (sp,obs,x1,x2 for dep; sp,obs,x11,x10,x01 for indep)
  #'@param type "dependent" or "independent"
  #'
  #'@return list containing max log-like, estimates of p, std.err of p,
  #'          gradient, hessian, completion code, num-iterations, model name,
  #'          analys type, design matrix, estimated N and std.err,
  #'          AIC, G-stat.
  #'
  #'@export
  #'
  #'@author Jim Hines
  #'
  #'@seealso dobserv_dep_obs
  #'

  cat('model:p(',gsub(' ','',as.character(frm)),')\n',sep='')
  dm=model.matrix(frm,mx); th.init=matrix(2,ncol(dm),1); nreal=nrow(mx)
  like=like_dep; ii=2; if (type!='dependent') { ii=3; like=like_indep}
  t1a=system.time(m1a<-nlm(like,p=th.init,d=d,dm=dm,hessian=T,iterlim=500))
  m1a=m1b=list(minimum=1e22);
  #try(t1b<-system.time(m1b<-nlm(like,p=th.init-th.init,d=d,dm=dm,hessian=T,iterlim=500)),silent=T)
  ier<-try(m1b<-nlm(like,p=th.init-th.init,d=d,dm=dm,hessian=T,iterlim=500),silent=TRUE)
  if (class(ier)[1] == "try-error") cat("ier=",ier,"\n")
  if (m1b$minimum<m1a$minimum) m1a=m1b
  m1a$name=paste0('p(',as.character(frm[-1]),')',collapse=''); m1a$type=type
  m1a$dm=dm; vcbeta=matrix(NA,length(m1a$estimate),length(m1a$estimate))
  res=try(vcbeta<-invhess(m1a$hessian,m1a$estimate),silent=TRUE); #res=try(vcbeta<-solve(m1a$hessian),TRUE);
  nbeta=ncol(vcbeta); grad=matrix(0,nreal,nbeta); grad2=gradN=matrix(0,nreal/2,nbeta); gradTN=matrix(0,1,nbeta)
  beta1=m1a$estimate; p=plogis(dm %*% beta1);
  tmp=matrix(p,ncol=2,byrow=T); pstar=tmp[,1]+tmp[,2]-tmp[,1]*tmp[,2]
  i=grep("^[XY]",colnames(d)); dd=d[,i]
  xdotdot=sum(dd)
  if (nrow(d)>1) xdotdot=rowSums(dd);
  N=xdotdot/pstar; totN=sum(N); eps1=.1e-12
  for (k in 1:nbeta) {
    beta1[k]=beta1[k]+eps1; p2=plogis(dm %*% beta1);
    tmp=matrix(p2,ncol=2,byrow=T); pstar2=tmp[,1]+tmp[,2]-tmp[,1]*tmp[,2]; N2=xdotdot/pstar2; totN2=sum(N2)
    grad[,k]=(p2-p)/eps1; grad2[,k]=(pstar2-pstar)/eps1; gradN[,k]=(N2-N)/eps1; gradTN[k]=(totN2-totN)/eps1
    beta1[k]=beta1[k]-eps1
  }
  vcp=matrix(NA,nreal,nreal); vcpstar=vcN=matrix(NA,nreal/2,nreal/2)
  if (class(res)[1]=='try-error')  {
    cat('*********************\nerror w/ hessian for model p(',as.character(frm),')\n************************\n',sep='')
  } else  {
    vcp=grad %*% vcbeta %*% t(grad); vcpstar=grad2 %*% vcbeta %*% t(grad2)
    vcN=gradN %*% vcbeta %*% t(gradN); vcTN=gradTN %*% vcbeta %*% t(gradTN)
  }
  varp=diag(vcp); sep=sqrt(varp); sepstar=sqrt(diag(vcpstar));
  m1a$p=p; m1a$sep=sep;
  rownames(m1a$p)=paste0(rep(d$sp,each=2),rep(1:2,nrow(d)))
  nhat=xdotdot/pstar; seTN=sqrt(vcTN)
  m1a$totN=totN; m1a$seTN=seTN
  varnhat=diag(vcN) + xdotdot*(1-pstar)/pstar^2; senhat = sqrt(varnhat)
  m1a$res=res; m1a$cputime=t1a; m1a$vcbeta=vcbeta; m1a$loglik=m1a$minimum; m1a$aic=2*m1a$minimum+2*ncol(m1a$dm)
  p=matrix(p,ncol=2,byrow=T); sep=matrix(sep,ncol=2,byrow=T)
  f0 = nhat - xdotdot; f0[f0==0]=.1e-22
  c = rep(.1e-22,length(varnhat)); i=which(varnhat>0)
  res=try(c[i] <- exp(1.96 * sqrt(log(1 + varnhat[i] / f0[i]^2))),TRUE)
  CI1 = xdotdot+f0 / c; CI2 = xdotdot+f0*c; CI2[CI2<CI1]=CI1[CI2<CI1]
  z=cbind(xdotdot,p[,1],sep[,1],p[,2],sep[,2],pstar,sepstar,nhat,senhat,CI1,CI2)
  colnames(z)=c('X..','p1','se(p1)','p2','se(p2)','p*','se(p*)','N','se(N)','lowCI','hiCI')
  rownames(z)=d$sp;  m1a$z=z
  m1a$ntotal=compute_var_sum_N(rep(1,nrow(d)),m1a,dd,rep(1,nrow(d)))
  names(m1a$ntotal)=c('X..','p*','se(p*)','N','se(N)','lowCI','hiCI')
  return(m1a)
}
compute_var_sum_N <- function(specx,m,d,speclist) {
  theta=m$estimate; th=plogis(m$dm %*% theta); nreal=nrow(m$dm)
  thj=th[seq(1,nreal,2)]; thjp1=th[seq(2,nreal,2)]; denom=thj+thjp1-thj*thjp1;
  k=which(speclist==specx); if (length(k)==1) x=sum(d[k,]) else x=rowSums(d[k,])  # x=x.. for each species
  p=denom[k]; N=sum(x/p); pstar=sum(x)/N; dndth=dpdth=NULL;
  tmp=sum(x*(1-p)/p^2)
  for (j in 1:length(theta)) {
    theta[j]=theta[j]+1e-12
    th=plogis(m$dm %*% theta); thj=th[seq(1,nreal,2)]; thjp1=th[seq(2,nreal,2)]; denom=thj+thjp1-thj*thjp1;
    p1=denom[k]; N1=sum(x/p1); dndth=cbind(dndth,(N1-N)/1e-12);
    pstar1=sum(x)/N1; dpdth=cbind(dpdth,(pstar1-pstar)/1e-12)
    theta[j]=theta[j]-1e-12
  }
  varn=dndth %*% m$vcbeta %*% t(dndth); sumx=sum(x); varnhat=varn+tmp
  varp=dpdth %*% m$vcbeta %*% t(dpdth)
  f0 = N - sumx; if (f0==0) f0=.1e-22
  c <- exp(1.96 * sqrt(log(1 + varnhat / f0^2)))
  CI1 = sumx+f0 / c; CI2 = sumx+f0*c
  seN=sqrt(varn+tmp)
  #if (sum(p<1e-5)>0) N=seN=CI1=CI2=NA
  return(c(sum(x),pstar,sqrt(varp),N,seN,CI1,CI2))
}
print_by_spec <- function(m,d) {
  #'  print estimates of p, summarized by species
  #'
  #'  \code{print_by_species} returns table of estimates, summarized by species
  #'  (or some other variable)
  #'
  #'@param m list with model output
  #'@param d data matrix of detections
  #'
  #'@return table of estimates of N, p, by species
  #'@export
  #'
  #'@author Jim Hines
  #'
  #'@examples
  #'\dontrun{
  #'  fname=file('sample.csv',package=dobserv)
  #'  z1=dobserv_dep_obs(fname,model=~obs);
  #'  z2=dobserv_dep_obs(fname,model=~-1+sp)
  #'  z3=dobserv_dep_obs(fname,model=~sp+obs)
  #'  z4=dobserv_dep_obs(fname,model=~sp*obs)
  #'  m=z1$m$"p(sp * obs)"
  #'  cat('\nN by TRT\n');
  #'  n1=print_by_spec(m,as.character(z1$data$SPECIES),  z1$d)
  #'  }
  specx=levels(as.factor(d$sp)); speclist=as.character(d$sp)
  if (is.null(d$X10)) d=cbind(d$X1,d$X01,0) else d=cbind(d$X10,d$X01,d$X11)
  n=round(t(sapply(specx,compute_var_sum_N,m,d,speclist)),4);
  n[,4:7]=round(n[,4:7],2);
  colnames(n)=c('X..','p*','se(p*)','N','se(N)','low95%CI','hi95%CI')
  i=which(n[,2]<1e-9); n[i,4:7]=NA
  return(n)
}
dobserv_dep_obs <- function(dobs_data, model=~1, obscov=NULL) {
  #'  runs double-observer analysis for dependent observers
  #'
  #'  \code{dobserv_dep_obs} returns model beta and real estimates from model
  #'
  #'@param dobs_data data frame with the following fields:
  #'\itemize{
  #'  \item \code{obs} observer number for primary observer
  #'  \item \code{sp} name of species
  #'  \item \code{X1} number of individuals of species detected by primary observer
  #'  \item \code{X01} number of individuals of species detected by secondary
  #'    observer that were missed by primary observer
  #'  }
  #'@param model model definition. Model definitions are in GLM form.
  #'  \itemize{
  #'  \item \code{~1} means detection is constant for all observers and species.
  #'  \item \code{~sp} means detection varies by species (constant across observers)
  #'  \item \code{~obs} means detection varies by observer (constant across species)
  #'  \item \code{~sp*obs} means detection varies by species and observer
  #'  \item \code{~xxxx} means detection varies by input covariate, xxxx
  #'}
  #'@param obscov Covariates which differ by observer.
  #'
  #'@note Covariates which are the same for both observers should be included
  #' in the input data-frame.  Covariates which differ by observer
  #' should be included as a separate data-frame (with twice the
  #' number of rows as the input data).
  #'@return list containing max log-like, estimates of p, std.err of p,
  #'          gradient, hessian, completion code, num-iterations, model name,
  #'          analys type, design matrix, estimated N and std.err,
  #'          AIC, G-stat.
  #'@export
  #'
  #'@author Jim Hines
  #'
  #'@seealso dobserv_indep_obs
  #'
  #'@examples
  #'#
  #'fname=system.file('extdata','sample.csv',package='dobserv')

  #'d1=read.csv(fname,as.is=TRUE)
  #'data=pool_lt10(d1);
  #'#   pool CAGO,PUMA even though > 10 detections because all detections were by same obs.
  #'i=which(data$sp=="CAGO" | data$sp=="PUMA"); data$sp[i]='grp1'
  #'
  #'#  summarize by species...
  #'sp=as.factor(data$sp); usp=levels(sp); nsp=length(usp); d2=matrix(0,2*nsp,2)
  #'jj=(as.numeric(sp)-1)*2+(data[,1]==data[1,1])+1
  #'for (i in 1:nrow(data)) d2[jj[i],]=d2[jj[i],]+as.numeric(data[i,3:4])
  #'d2=data.frame(obs=rep(1:2,nsp),sp=rep(as.character(usp),each=2),X1=d2[,1],X01=d2[,2])
  #'
  #'  #   run some models...
  #'    #    ~obs means detection depends only on observer,
  #'    #    ~sp means detection depends only on species,
  #'    #    ~xxxx means detection depends on input covariate xxxx
  #'
  #'z1=dobserv_dep_obs(d2,model=~sp*obs)
  #'#   z1 is a list with output from the model above
  #'
  #'#  print results from model, p(sp*obs) (ie., detection different for each
  #'#                                         species and observer)
  #'print(round(z1$z,4))
  #'
  d=dobs_data[,colnames(dobs_data) %in% c('X1','X01')]
  nr=nrow(d);    # no. real params = 2*(# sites) ... p(obs1),p(obs2) for each site
  mx=data.frame(int=rep(1,nr*2), obs=rep(0:1,nr), sp=as.factor(rep(dobs_data$sp,each=2)))
  rownames(mx)=NULL

  if (ncol(dobs_data)>=5) {
    for (i in 5:ncol(dobs_data)) mx=cbind(mx,rep(dobs_data[,i],each=2))
    colnames(mx)[4:ncol(mx)]=colnames(dobs_data)[-1:-4]
  }
  if (!is.null(obscov)) mx=cbind(mx,obscov)

  ##          run a model
    m=runmod(model,mx,d); dd=cbind(d$X1,d$X01)
    j=which(dd>0,arr.ind=T);
    m$G=sum(2*log(dd[j]/(expval[j]+1e-9))*dd[j])

  return(m)
}

dobserv_indep_obs <- function(dobsdata,model=~obs, obscov=NULL) {
  #'  runs double-observer analysis for independent observers
  #'
  #'  \code{dobserv_indep_obs} returns aic table of results
  #'
  #'@param dobsdata data frame with the following fields:
  #'\itemize{
  #'  \item \code{obs} observer number for primary observer
  #'  \item \code{species} name of species
  #'  \item \code{X10} number of individuals of species detected by observer 1
  #'  and missed by observer 2
  #'  \item \code{X01} number of individuals of species missed by observer 1
  #'  and detected by observer 2
  #'  \item \code{X11} number of individuals of species detected by both observers
  #'  }
  #'@param model list of model definitions. Model definitions are in GLM form.
  #'  \itemize{
  #'  \item \code{~1} means detection is constant for all observers and species.
  #'  \item \code{~sp} means detection varies by species (constant across observers)
  #'  \item \code{~obs} means detection varies by observer (constant across species)
  #'  \item \code{~sp*obs} means detection varies by species and observer
  #'  \item \code{~xxxx} means detection varies by input covariate, xxxx
  #'}
  #'@param obscov Covariates which differ by observer.
  #'
  #'@note Covariates which are the same for both observers should be included
  #' in the input data-frame.  Covariates which differ by observer
  #' should be included as a separate data-frame (with twice the
  #' number of rows as the input data).
  #'
  #'@return list containing max log-like, estimates of p, std.err of p,
  #'          gradient, hessian, completion code, num-iterations, model name,
  #'          analys type, design matrix, estimated N and std.err,
  #'          AIC, G-stat.
  #'@export
  #'
  #'@author Jim Hines
  #'
  #'@seealso dobserv_dep_obs
  #'
  #'@examples
  #'#
  #' \dontrun{
  #'
  #'fname=system.file('extdata','sample3.csv',package='dobserv')
  #'
  #'  #  read sample data file which includes a covariate (temperature) in the last column
  #'d3=read.csv(fname,header=F,skip=1); colnames(d3)=c('obs','sp','X10','X01','X11','temp')
  #'  #  run a few models...
  #'z1=dobserv_indep_obs(d3,model=~obs)
  #'z2=dobserv_indep_obs(d3,model=~temp*obs)
  #'  #  print estimates from 2nd model for each species...
  #'print(round(z2$z,4))
  #'  #  print estimate of total number of individuals...
  #'print(round(z2$ntotal,4))
  #'  }
  d=dobsdata[,colnames(dobsdata) %in% c('X10','X01','X11')]
  nr=nrow(dobsdata)
  mx=data.frame(int=rep(1,nr*2), obs=rep(0:1,nr), sp=as.factor(rep(dobsdata$sp,each=2)))

  if (ncol(dobsdata)>=6) {
    for (i in 6:ncol(dobsdata)) mx=cbind(mx,rep(dobsdata[,i],each=2))
    colnames(mx)[4:ncol(mx)]=colnames(dobsdata)[-1:-5]
  }
  if (!is.null(obscov)) mx=cbind(mx,obscov)
  rv=try(mm<-runmod(model,mx,d,'independent'))
  if (!class(rv)[1]=="try-error") {
    m=mm
    j=which(d>0,arr.ind=T);
    m$G=sum(2*log(d[j]/(expval[j]+1e-9))*d[j])
  }
  return(m)
}

aictbl <- function(m,d) {
  #'  computes AIC table from model-list, m
  #'
  #'  \code{aictbl} returns aic table of results
  #'
  #'@param m list of models
  #'@param d input data-frame
  #'
  #'@return aic table of model results
  #'@export
  #'@author Jim Hines
  #'
  #'@seealso dobserv_indep_obs
  #'
  #'@examples
  #'#
  #' \dontrun{
  #'  }
  aictbl=NULL; dd=cbind(d$X1,d$X01,d$Y1,d$Y01)
  ness=sum(dd); maxpar=j=0; K=sapply(1:length(m),function(i)length(m[[i]]$estimate))
  for (i in 1:length(m)) if (K[i]>maxpar) {maxpar=K[i]; j=i}
  chat=1; Gdf=(2*nrow(d)-(maxpar-1)+1);
  if (m[[j]]$G>0 & Gdf>0) chat=m[[j]]$G/Gdf;  llconst=0; if (chat<1) chat=1
  if (m[[1]]$type=="dependent") llconst=sum(lgamma(rowSums(dd[,1:2])+1))-sum(lgamma(dd+1))
  if (m[[1]]$type=="independent") llconst=sum(lgamma(rowSums(dd)+1))-sum(lgamma(dd+1))
  cat('c-hat=',chat,'ness=',ness,'llconst=',llconst,'\n');
  for (i in 1:length(m)) {
    loglik=m[[i]]$minimum-llconst
    m[[i]]$aic=2*loglik+2*K[i]
    m[[i]]$aicc=2*loglik + 2*K[i] + 2*K[i]*(K[i] + 1)/(ness - K[i] - 1)
    m[[i]]$qaicc=-2*loglik/chat + 2*K[i] + 2*K[i]*(K[i] + 1)/(ness - K[i] - 1)
    aictbl=rbind(aictbl,cbind(m[[i]]$name,K[i],round(loglik,4),round(m[[i]]$aicc,4)))
  }
  aic=as.numeric(aictbl[,4]); dlta=aic-min(aic); tmp=exp(-dlta/2); aicwgt=tmp/sum(tmp)
  aictbl=cbind(aictbl,round(dlta,4),round(aicwgt,4))
  colnames(aictbl)=c('name','npar','negLogLike','AICc','deltaAICc','AICcwgt')
  if (chat>1) colnames(aictbl)=gsub("AIC","QAIC",colnames(aictbl))
  o=order(as.numeric(aictbl[,5]))
  return(list(table=aictbl[o,],chat=chat))
}

